/*
 * @(#)ContentModel.java	1.3 98/09/02
 * 
 * Copyright (c) 1994-1998 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package com.sun.xml.parser;

import java.util.Vector;
import java.util.Enumeration;
import java.io.*;


/**
 * A representation of a "children" content model.  These are basically a
 * regular expression; other content models are simpler.  There is an
 * SGML compatibility restriction on DTDs that such content models be
 * deterministic, which in this sense just means that backtracking isn't
 * needed to validate against it.
 *
 * <P> At the moment, for expediency, nondeterministic models are neither
 * tested for nor are they handled reasonably.  This could be done after
 * each element's content model is fully parsed.
 *
 * @version 	1.3 (via HotJava, 1.14, 12/22/97)
 *
 * @author Arthur van Hoff
 * @author David Brownell
 */
final class ContentModel
{
    /**
     * Type. Either '*', '?', '+'; or connectives ',', '|'; or
     * zero for content that's an element.
     */
    public char type;

    /**
     * The content. Either an Element name, or a ContentModel.
     */
    public Object content;

    /**
     * The next content model (in a ',' or '|' connective expression).
     * "next" has a list of connectives of the same type.
     */
    public ContentModel next;


    /**
     * Create a content model for an element.
     */
    public ContentModel (String element) {
	this.type = 0;
	this.content = element;
    }

    /**
     * Create a content model of a particular type.
     * Normally used to specify a frequency, or to start a connective.
     */
    public ContentModel (char type, ContentModel content) {
	this.type = type;
	this.content = content;
    }

    /**
     * Return true if the content model could
     * match an empty input stream.
     */
    public boolean empty () {
	switch (type) {
	  case '*':
	  case '?':
	    return true;

	  case '+':
	  case 0:
	    return false;

	  case '|':
	    if (content instanceof ContentModel
		    && ((ContentModel)content).empty ()) {
		return true;
	    }
	    for (ContentModel m = (ContentModel)next;
		    m != null;
		    m = m.next) {
		if (m.empty ())
		    return true;
	    }
	    return false;

	  case ',':
	    if (content instanceof ContentModel) {
		if (!((ContentModel)content).empty ()) {
		    return false;
		}
	    } else {
		return false;
	    }
	    for (ContentModel m = (ContentModel)next;
		    m != null;
		    m = m.next) {
		if (!m.empty ())
		    return false;
	    }
	    return true;

	  default:
	    throw new InternalError ();
	}
    }

    // this benefits notably from caching this
    // function ... tokens map to { t, f, unknown } and
    // "unknown" triggers finding  t/f and caches the result.

    // HJB also assigns elements an index, speeds up such checks.

    /**
     * Return true if the token could potentially be the
     * first token in the input stream.
     */
    public boolean first (String token) {
	switch (type) {
	  case '*':
	  case '?':
	  case '+':
	  case 0:
	    if (content instanceof String)
		return content == token;
	    else
		return ((ContentModel)content).first (token);

	  case ',':
	    if (content instanceof String && content == token)
		return true;
	    if (((ContentModel)content).first (token))
		return true;
	    if (!((ContentModel)content).empty ())
		return false;
	    for (ContentModel m = (ContentModel)next;
		    m != null;
		    m = m.next) {
		if (m.content instanceof String && m.content == token)
		    return true;
		if (((ContentModel)m.content).first (token))
		    return true;
		if (!((ContentModel)content).empty ())
		    return false;
	    }
	    return false;

	  case '|':
// XXX deleted val/valSet caches here ...
	    if (content instanceof String && content == token)
		return true;
	    if (((ContentModel)content).first (token))
		return true;
	    for (ContentModel m = (ContentModel)next;
		    m != null;
		    m = m.next) {
		if (m.content instanceof String && m.content == token)
		    return true;
		if (((ContentModel)m.content).first (token))
		    return true;
	    }
	    return false;

	  default:
	    throw new InternalError ();
	}
    }

    /**
     * Return the element that must be next.
     */
    public String next () {
	switch (type) {
	  case '|':
	  case '*':
	  case '?':
	    return null;

	  case 0:
	    return (String) content;

	  case '+':
	  case ',':
	    if (content instanceof String)
		return (String) content;
	    else
		return ((ContentModel)content).next ();

	  default:
	    throw new InternalError ();
	}
    }

    /**
     * Convert to a string (for debugging).
     *
    public String toString () {
	return toString (true);
    }

    private String contentString ()
    {
	if (content instanceof ContentModel)
	    return ((ContentModel)content).toString (false);
	else
	    return (String) content;
    }

    private String toString (boolean isOuter)
    {
	String	temp = contentString ();

	switch (type) {
	  case '*':
	  case '?':
	  case '+':
	    if (isOuter && temp.charAt (0) != '(')
		return "(" + temp + type + ")";
	    else
		return temp + type;

	  case 0:
	    if (isOuter)
		return "(" + temp + ")";
	    else
		return temp;

	  case ',':
	  case '|':
	    if (next == null)
		return temp;
	    for (ContentModel m = next; m != null; m = m.next)
		temp += type + m.contentString ();
	    return "(" + temp + ")";

	  default:
	    throw new InternalError ("foo");
	}
    }
    /**/
}
